\ bmwslat.fs   version 0.0.2  (c) 20 January, 1999,
\  Bruce R. McFarling, ecbm@cc.newcastle.edu.au
\  all rights reserved.

\ This source is free for distribution using any
\ electronic or magnetic storage medium, and free
\ for any educational OR non-commercial use.

GET-ORDER \ The file must end with SET-ORDER
ONLY FORTH TOOLS DEFINITIONS

MARKER *BMWSLAT*
VOCABULARY BMWSLAT
ALSO BMWSLAT DEFINITIONS

\ This file uses BMWTOOLS.FS
\ A collection of tools used in the BMW system
S" bmwtools.fs" REQUIRED TOOLS BMWTOOLS

\ A SLAT is a collection of rows the available width of the display
\ The operation SLAT returns the absolute row which
\ is row 0 for slat operations

\ * SLAT values & constants
\ Modification note:
\  You may wish to change the constants into values
\     so that display parameters may be query driven,
\     or stored in a configuration file.
\  If so, make sure that a using word cannot set them with TO
\     SLAT and PLANKS shows how to export an external parameter

0 CONSTANT Top-Row               \ The upper limit for slats

\ Slat-Value is for the use of the display system
\ SLAT is for the information of applications
\ Applications modify Slat-Value using slat operations
\ Applications may not depend on the existence of Slat-Value,
\ nor its base or units, if it exists.

\ * SLAT is the row offset for SLAT-XY -- IMPORTED AS A PARAMETER
Top-Row VALUE Slat-Value
: SLAT ( -- slat-origin ) Slat-Value ;

\ Gforth for DOS runs into the DOS autoscroll at column 80
\ Use display max if you can emit to last col w/out scrolling
\ as in a Linux virtual console or rxvt, display-max -1 if
\ you run into an auto-scrolling terminal

79 CONSTANT SLAT-WIDTH

\ Use 43 for one of the taller EGA text modes
\ Use 24 for VT-100,
\ 25 for most common MS-DOS text mode & Linux Vconsole
24 CONSTANT MAX-PLANKS

MAX-PLANKS CONSTANT DEFAULT-PLANKS
MAX-PLANKS 4 / VALUE SOME-PLANKS

\ * PARAMETER: users modify PLANKS with slat operations
\ SLAT MODEL: Always start plank-value at default
DEFAULT-PLANKS VALUE Planks-Value
: PLANKS ( -- planks ) Planks-Value ;

\ * the top plank is 0 and the last plank is MAX-1
: TOP-PLANK ( -- plank# ) 0 ; 
: END-PLANK ( -- plank# ) PLANKS 1- ;

\ If you have the means to query the dimensions of the display
\ then add the code here
\ If the display dimensions have changed, return TRUE
: DIMENSIONS? ( -- col plank flag ) SLAT-WIDTH PLANKS 0 ;

\ ** SLAT operations

\ * Column u1, row u2: row is relative to the top of the slat
: SLAT-XY ( col plank -- ) Slat-Value + AT-XY ;

\ * column 0 of plank
: SLAT-Y ( plank -- ) 0 SWAP SLAT-XY ;

\ * clear single row relative to slat -- which is a "plank"
\ leaves cursor at start of the plank; clipped at end-plank
: CLEAR-PLANK ( plank -- )
   END-PLANK UMIN
   DUP SLAT-Y SLAT-WIDTH SPACES
   SLAT-Y
;

\ * clear plank from column to end of plank
: CLEAR-PLANK>>   ( col plank -- )
   2DUP DUP Planks-Value > IF
      TUCK SLAT-XY SLAT-WIDTH SWAP - SPACES
   ELSE DROP
   THEN SLAT-XY
;

\ * clear the display area of the slat
: CLEAR-SLAT ( -- )
   Planks-Value 0 ?DO I CLEAR-PLANK LOOP 0 SLAT-Y 
;

\ * clear remaining planks in slat
\    implementation dependent definition should be faster
: CLEAR-SLAT>> ( plank -- )
   DUP Planks-Value U< IF
      Planks-Value OVER ?DO I CLEAR-PLANK LOOP
      SLAT-Y
   ELSE DROP
   THEN
;

\ * Read existing slat values
\ NB. While ( col plank ) stack pictures are two dimensions
\   ( slat plank ) stack pictures are base+extent
: SLAT> ( -- slat planks ) Slat-Value Planks-Value ;

\ * Set-up a new slat origin and height
\ The height is forced within the min and max planks
\ If the settings are below the end of screen,
\  the top row of the slat is shifted up
\ The cursor is placed at the beginning of plank 0

\ Shift the base up if the extent overflows
: SLAT-RETREAT-UP ( slat planks max-planks -- slat' planks )
   DUP-3RD DUP-3RD +          \ sl pl Max[sl+pl] sl+pl
   2DUP < IF - UNDER+ ELSE 2DROP THEN
;

: >SLAT ( slat planks -- )
   ?DUP 0= IF SOME-PLANKS THEN    \ 'few' is the default
   1 MAX-PLANKS CLAMP             \ . min<=plank<=max
   MAX-PLANKS SLAT-RETREAT-UP
   TO Planks-Value   TO Slat-Value
;


\ * Split slat into two: both must be at least one plank high
: SPLIT-SLAT ( slat planks top-planks -- sl2 pl2 sl1 pl1 )
   OVER 1- UMIN 1 UMAX        \ slat planks pl1
   TUCK TUCK - >R             \ sl1 pl1 pl1     R: pl2
   DUP-3RD + R> 2SWAP
; 

\ * If there are no planks, two empty slats are return
: PRY-FIRST ( slat planks|0 -- sl+ pl- sl 1 | sl 0 sl 0 )
   DUP 0= IF 2DUP EXIT THEN
   1- SWAP DUP 1+ -ROT 1
;

\ * If there are no planks, two empy slats are returned
: PRY-LAST ( slat planks|0 -- sl pl- sl+pl-1 1 | sl 0 sl 0 )
   DUP 0= IF 2DUP EXIT THEN
   1- 2DUP + 1
;

\ * Let user store and restore slats
\ COMPILES: ... ( -- ) SLAT-VARIABLE <name> ...
: SLAT-VARIABLE ( DOES: -- addr )   2VARIABLE ;
: SLAT! ( slat planks addr -- )     2! ;
: SLAT@ ( addr -- slat plank )      2@ ;
: @>SLAT ( addr -- )                2@ >SLAT ;
: EMPTY-SLAT! ( addr -- )           0 0 ROT 2! ;
: NEW-SLAT ( slat-addr -- )         SLAT@ >SLAT CLEAR-SLAT ;
: COPY-SLAT ( slat planks addr -- slat planks )
   >R 2DUP R> SLAT!
;


CR .( BMWSLAT available in current TOOLS vocabulary )

SET-ORDER
